# -*- coding: utf-8 -*-

# This information is located in its own file so that it can be loaded
# without importing the main package when its dependencies are not installed.
# See: https://packaging.python.org/guides/single-sourcing-package-version

__author__ = """David Huard"""
__email__ = "huard.david@ouranos.ca"
__version__ = "0.5.0"
